/*
	mactioneditsdialog.cpp - action edit dialog
	Copyright (C) 2003  Konrad Twardowski <kdtonline@poczta.onet.pl>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "mactioneditdialog.h"
#include "mtip.h"

#include <qlayout.h>

#include <klineedit.h>
#include <klocale.h>

/* constructor */

MActionEditDialog::MActionEditDialog(QWidget *parent, const char *name)
	: KDialogBase(
		Plain,
		i18n("Edit"),
// TODO: help
		Ok | Cancel,
		Ok, // default button
		parent,
		name,
		TRUE, // modal
		TRUE // separator
		),
	// init
	userCommand("")
{
	QFrame *mainWidget = plainPage();
	QVBoxLayout *mainLayout = new QVBoxLayout(mainWidget, 5, 5);

	// action name
	l_what = new QLabel(mainWidget, "QLabel::l_what");
	l_what->setAlignment(AlignCenter);
	mainLayout->addWidget(l_what);

	// method label
	QLabel *l_method = new QLabel(i18n("&Method"), mainWidget, "QLabel::l_method");
	mainLayout->addWidget(l_method);

	// method combo box
	methods = new QComboBox(mainWidget, "QComboBox::method");
	l_method->setBuddy(methods);
	connect(
		methods, SIGNAL(activated(int)),
		this, SLOT(methodsChangeSlot(int)));
	mainLayout->addWidget(methods);

	// command label
	l_command = new QLabel(i18n("Comm&and"), mainWidget, "QLabel::l_command");
	mainLayout->addWidget(l_command);

	// command edit line
	command = new KLineEdit(mainWidget, "KLineEdit::command");
	l_command->setBuddy(command);
	mainLayout->addWidget(command);

	// -
	mainLayout->addStretch();

	// tip
	MTip *tip = new MTip(MTip::TipType_Warning, mainWidget, "MTip::tip");
	tip->setTipText(
		i18n("In most cases you need privileges to shut down system (e.g. run /sbin/reboot or /sbin/shutdown)") + "<br><br>" +
		i18n("Manuals:") +
		"<ul>" \
		"<li><a href=\"man:/halt\">halt, poweroff, reboot</a></li>" \
		"<li><a href=\"man:/shutdown\">shutdown</a></li>" \
		"<li><a href=\"man:/xlock\">xlock</a></li>" \
		"</ul>"
	);
	mainLayout->addWidget(tip);

	// focus combo box
	methods->setFocus();
}

/* destructor */

MActionEditDialog::~MActionEditDialog()
{
}

/* public */

void MActionEditDialog::setMethod(const MMainWindow::Method value) const
{
	methods->setCurrentItem(value);
	l_command->setEnabled(value == MMainWindow::Method_UserCommand);
	command->setEnabled(value == MMainWindow::Method_UserCommand);
	switch (value)
	{
		case MMainWindow::Method_KDE:
			command->setText("");
			break;
		case MMainWindow::Method_DefaultCommand:
			command->setText(methods->text(MMainWindow::Method_DefaultCommand));
			break;
		case MMainWindow::Method_UserCommand:
			command->setText(userCommand);
			break;
	}
}

void MActionEditDialog::setWhat(const MMainWindow::What value)
{
	l_what->setText(mainWindow->getActionName(value));
}

/* private slots */

void MActionEditDialog::methodsChangeSlot(int index)
{
	if (command->isEnabled())
		userCommand = command->text();
	setMethod((MMainWindow::Method)index);
}
