/*
	mmainwindow.cpp - main window
	Copyright (C) 2003  Konrad Twardowski <kdtonline@poczta.onet.pl>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// TODO: this file is to large!

#ifndef TEST_KSHUTDOWN_DCOP_API
//	#define TEST_KSHUTDOWN_DCOP_API
// uncomment above line to test kshutdowndcopapi.*
// see also "taskTest"
#endif // TEST_KSHUTDOWN_DCOP_API

#include "appscheduler.h"
#include "extras.h"
#include "mmainwindow.h"
#include "mmessagedialog.h"
#include "mradiobutton.h"
#include "mschedulertab.h"
#include "msettingsdialog.h"
#include "mstatstab.h"
#include "msystemtray.h"
#include "mwizard.h"

#ifdef TEST_KSHUTDOWN_DCOP_API
#include "kshutdowndcopapi.h"
#endif // TEST_KSHUTDOWN_DCOP_API

#include <qlayout.h>
#include <qtimer.h>
#include <qtooltip.h>
#include <qwhatsthis.h>

#include <dcopclient.h>
#include <kaction.h>
#include <kaudioplayer.h>
#include <kconfig.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <krun.h>
//#include <kstandarddirs.h>
#include <kwin.h>

/* globals */

#ifdef TEST_KSHUTDOWN_DCOP_API
KShutDownDCOPAPI *taskTest = 0;
#endif // TEST_KSHUTDOWN_DCOP_API

MMainWindow *MMainWindow::mainApplicationWindow = 0;

/* constructor */

MMainWindow::MMainWindow(QWidget *parent, const char *name)
	: KMainWindow(parent, name),
	// init
	_active(FALSE),
	_cancelled(FALSE),
	_testMode(FALSE),
	totalExit(FALSE),
	delayType(DelayType_Seconds),
	margin(4),
	spacing(4),
	oldSec(-1),
	_modalMessageDialog(0),
	what(What_Nothing),
	when(When_Now)
{
	mainApplicationWindow = this;

	connect(
		kapp, SIGNAL(aboutToQuit()),
		this, SLOT(aboutToQuitSlot()));

	// kdDebug() << "MMainWindow::MMainWindow()" << endl;

	// init date/time
	dt_end = new QDateTime();
	dt_start = new QDateTime();

	// init "extras" manager
	extras = new Extras(this, "Extras::extras");

	// action timeout timer
	timer = new QTimer(this);
	connect(
		timer, SIGNAL(timeout()),
		this, SLOT(checkTimeSlot()));

	// main widget
	QWidget *mainWidget = new QWidget(this, "QWidget::mainWidget");
	setCentralWidget(mainWidget);
	// horizontal
	QHBoxLayout *mainLayout = new QHBoxLayout(mainWidget);
	mainLayout->setMargin(margin);
	mainLayout->setSpacing(spacing);

	// vertical (left) - tabs + bottom buttons
	QWidget *leftWidget = new QWidget(mainWidget, "QWidget::leftWidget");
	QVBoxLayout *leftLayout = new QVBoxLayout(leftWidget);

	// add tabs
	tabWidget = new QTabWidget(leftWidget, "QTabWidget::tabWidget");
	// time tab
	timeTab = initTimeTab(leftWidget);
	tabWidget->addTab(timeTab, SmallIcon("clock"), i18n("&Time"));
	// scheduler tab
	schedulerTab = new MSchedulerTab(tabWidget, "MSchedulerTab::schedulerTab");
	tabWidget->addTab(schedulerTab, kapp->miniIcon(), "");
	// statistics tab
	statsTab = new MStatsTab(tabWidget, "MStatsTab::statsTab");
	tabWidget->addTab(statsTab, SmallIcon("misc"), i18n("Stat&istics"));

	connect(
		tabWidget, SIGNAL(currentChanged(QWidget *)),
		this, SLOT(tabChangedSlot(QWidget *)));

	tabWidget->setMargin(0);
	leftLayout->addWidget(tabWidget);

	// init bottom buttons
	leftLayout->addWidget(initBottomButtons(leftWidget));

	mainLayout->addWidget(leftWidget);

	// init main buttons (right)
	mainLayout->addWidget(initButtons(mainWidget));

	// init system tray
	systemTray = new MSystemTray(this, "MSystemTray::systemTray");

	// init DCOP
	DCOPClient *client = kapp->dcopClient();
	client->registerAs(kapp->name(), FALSE);

	// init scheduler
	_appScheduler = new AppScheduler();
	client->setDefaultObject(_appScheduler->objId()); // KShutdownIface

	readConfig();

	updateSchedulerTab();

#ifdef TEST_KSHUTDOWN_DCOP_API
	taskTest = new KShutDownDCOPAPI(client, "Hello, World!", ACTION_REBOOT);
	qDebug("taskTest->id(): %d", taskTest->id());
#endif // TEST_KSHUTDOWN_DCOP_API
}

/* destructor */

MMainWindow::~MMainWindow()
{
	// kdDebug() << "MMainWindow::~MMainWindow()" << endl;

#ifdef TEST_KSHUTDOWN_DCOP_API
	if (taskTest)
		delete taskTest;
#endif // TEST_KSHUTDOWN_DCOP_API

	delete _appScheduler;
	delete dt_end;
	delete dt_start;
	delete extras;
}

/* public */

void MMainWindow::cancelAction()
{
	setActive(What_Nothing, FALSE);
	setCancelled(TRUE);
	if (_modalMessageDialog)
		_modalMessageDialog->cancelAction();
}

bool MMainWindow::endSession(const KApplication::ShutdownType type, const What action)
{
	// kdDebug() << "MMainWindow::endSession()" << endl;
	totalExit = FALSE;

	// test mode
	if (testMode())
	{
		KMessageBox::information(this, getActionName(action), i18n("Test"));

		return TRUE;
	}

	totalExit = TRUE;
	if (!kapp->requestShutDown(KApplication::ShutdownConfirmNo, type))
	{
		KMessageBox::error(this,
			i18n("Could not logout properly.\nThe session manager cannot be contacted."));
		totalExit = FALSE;

		return FALSE; // error
	}

	return TRUE; // ok
}

bool MMainWindow::execAction(const What action, const bool stopTimer)
{
	if (stopTimer)
		setActive(action, FALSE);

	totalExit = FALSE;

	Method m = Method_KDE;
	QString c;

	switch (action)
	{
		// nothing
		case What_Nothing:
			return FALSE; // error

		// shut down
		case What_ShutDown:
			if (isRestricted("action_shutdown"))
				return FALSE; // error

			totalExit = TRUE;
			getMethod("Shut Down", m, c, DEFAULT_SHUT_DOWN_COMMAND);
			if (m == Method_KDE)
			{
				return endSession(
					KApplication::ShutdownTypeHalt,
					action);
			}
			break;

		// reboot
		case What_Reboot:
			if (isRestricted("action_reboot"))
				return FALSE; // error

			totalExit = TRUE;
			getMethod("Reboot", m, c, DEFAULT_REBOOT_COMMAND);
			if (m == Method_KDE)
			{
				return endSession(
					KApplication::ShutdownTypeReboot,
					action);
			}
			break;

		// lock screen
		case What_LockScreen:
			if (isRestricted("action_lockscreen"))
				return FALSE; // error

			getMethod("Lock Screen", m, c, DEFAULT_LOCK_SCREEN_COMMAND);
			if (m == Method_KDE)
			{
				// test mode
				if (testMode())
				{
					KMessageBox::information(
						this,
						getActionName(action),
						i18n("Test"));

					return TRUE; // ok
				}

				// this is a modified "void Lockout::lock()"
				// from "Lock/Logout Applet" (lockout.cpp [3.1.4])
				QCString kdesktop("kdesktop");
				int kshutdown_screen_number = qt_xscreen();
				if (kshutdown_screen_number)
					kdesktop.sprintf("kdesktop-screen-%d", kshutdown_screen_number);

				if (!kapp->dcopClient()->send(kdesktop, "KScreensaverIface", "lock()", ""))
				{
					KMessageBox::error(
						this,
						i18n("kdesktop: DCOP call failed!"));

					return FALSE; // error
				}
				hide(); // hide window

				return TRUE; // ok
			}
			break;

		// logout
		case What_Logout:
			if (isRestricted("action_logout"))
				return FALSE; // error

			totalExit = TRUE;
			getMethod("Logout", m, c, DEFAULT_LOGOUT_COMMAND);

			if (m == Method_KDE)
				return endSession(KApplication::ShutdownTypeNone, action);

			break;

		// "extras" action
		case What_Extras:
			if (isRestricted("action_extras"))
				return FALSE; // error

			// test mode
			if (testMode())
			{
				KMessageBox::information(
					this,
					extras->actionDescription() + "\n" +
					extras->fileToExecute(),
					i18n("Test"));

				return TRUE; // ok
			}

			return extras->execAction();

	}

	// test mode
	if (testMode())
	{
		KMessageBox::information(this,
			i18n("Command:") + " " + c, i18n("Test"));

		return TRUE; // ok
	}

	// run default or user command
	if (runCommand(c))
	{
		if (action == What_LockScreen)
			hide(); // hide window

		return TRUE; // ok
	}

	totalExit = FALSE;

	return FALSE; // error
}

QPixmap MMainWindow::getActionIcon(const What action) const
{
	switch (action)
	{
		case What_Nothing:
			return SmallIcon("misc");

		case What_ShutDown:
			return SmallIcon("exit");

		case What_Reboot:
			return SmallIcon("exit");

		case What_LockScreen:
			return SmallIcon("lock");

		case What_Logout:
			return SmallIcon("exit");

		case What_Extras:
			return SmallIcon("bookmark");
	}

	return SmallIcon("misc");
}

QString MMainWindow::getActionName(const What value) const
{
	switch (value)
	{

		case What_Nothing:
			return i18n("Nothing");

		case What_ShutDown:
			return i18n("Shut Down");

		case What_Reboot:
			return i18n("Reboot");

		case What_LockScreen:
			return i18n("Lock Screen");

		case What_Logout:
			return i18n("Logout");

		case What_Extras:
			return extras->getActionName();

		default:
			return i18n("Unknown");

	}
}

void MMainWindow::getMethod(const QString &configEntry, Method &method, QString &command, const QString &defaultCommand)
{
	KConfig *conf = config();

	if (!conf->hasGroup(configEntry))
	{
		method = Method_KDE;
		command = defaultCommand;

		return;
	}

	conf->setGroup(configEntry);

	// read method
	method = (Method)conf->readNumEntry("Method", Method_KDE);
	if (method < Method_KDE || method > Method_UserCommand)
		method = Method_KDE;

	switch (method)
	{
		case Method_KDE:
			command = conf->readEntry("Command", defaultCommand);
			break;
		case Method_DefaultCommand:
			command = defaultCommand;
			break;
		case Method_UserCommand:
			command = conf->readEntry("Command", defaultCommand);
			if (command == "")
				method = Method_KDE;
			break;
	}
}

bool MMainWindow::messageDialogAccepted(const bool modal, const What actionToExecute, const QString &text)
{
	if (!warningMessageEnabled())
		return TRUE; // accept

	// make main window visible
	show();
	KWin::setOnDesktop(winId(), KWin::currentDesktop());
	setFocus();

	// show message dialog
	MMessageDialog *messageDialog = new MMessageDialog(modal, warningMessageDelay, actionToExecute, text, this, "MMessageDialog::messageDialog");
	Q_CHECK_PTR(messageDialog);

	// modal
	if (modal)
	{
		_modalMessageDialog = messageDialog;
		if (_modalMessageDialog->exec() == QDialog::Accepted) // exec dialog
		{
			delete _modalMessageDialog;
			_modalMessageDialog = 0;

			return TRUE; // accepted
		}
		delete _modalMessageDialog;
		_modalMessageDialog = 0;

		return FALSE; // rejected
	}

	// non-modal
	messageDialog->show();

	return TRUE;
}

void MMainWindow::notifyUser(const int secs) const
{
	switch (secs)
	{
		case 300: // 5 minutes
			playSound("5m");
			tray()->popupMessage(MSystemTray::Reason_Warning, i18n("5 minutes warning"));
			break;
		case 60: // 1 minute
			playSound("1m");
			tray()->popupMessage(MSystemTray::Reason_Warning, i18n("1 minute warning"));
			break;
		case 10: // 10 seconds
			tray()->popupMessage(MSystemTray::Reason_Warning, i18n("10 seconds warning"));
			break;
		case 3: // 3 seconds
			playSound("three");
			break;
		case 2: // 2 seconds
			playSound("two");
			break;
		case 1: // 1 second
			playSound("one");
			break;
	}
}

void MMainWindow::play(const QString &dir, const QString &name) const
{
	QString s = dir;
	if (!s.endsWith("/"))
		s.append("/");
	s.append(name + ".ogg");
	// kdDebug() << "MMainWindow::play(): KAudioPlayer::play(`" << s << "')" << endl;
	KAudioPlayer::play(s);
}

bool MMainWindow::runWizard()
{
	if (isRestricted("action_wizard"))
		return FALSE;

	MWizard *wizard = new MWizard(this, "MWizard::wizard");
	bool retVal = wizard->runWizard();
	delete wizard;

	return retVal;
}

void MMainWindow::setActive(const What value, const bool yes)
{
	tray()->hidePopupMessages();

	setCancelled(FALSE);

	if (isActive() == yes)
	{
		// kdDebug() << "MMainWindow::setActive: (isActive() == yes)" << endl;
		return; // this should never happen, because buttons are disabled
	}

	if (yes && !confirm(value))
		return; // action not confirmed by user; do nothing

	// immediate action
	if (when == When_Now)
	{
		execAction(value);

		return;
	}

	_active = yes;
	what = value;

	// disable widgets
	if (_active)
	{
		cb_when->setEnabled(FALSE);
		b_misc->setEnabled(FALSE);
		sb_delay->setEnabled(FALSE);
		cb_delayType->setEnabled(FALSE);
		te_time->setEnabled(FALSE);
		de_date->setEnabled(FALSE);
	}
	// enable widgets
	else
	{
		setWhen(when); // this will enable widgets correctly
	}

	b_shutDown->setEnabled(!_active);
	b_reboot->setEnabled(!_active);
	b_logout->setEnabled(!_active);
	b_wizard->setEnabled(!_active);
	b_extras->setEnabled(!_active);

	// start time = now
	dt_start->setDate(QDate::currentDate());
	dt_start->setTime(QTime::currentTime());
	// end time = start time
	dt_end->setDate(dt_start->date());
	dt_end->setTime(dt_start->time());

	// round down msec
	int ms = dt_end->time().msec();
	if (ms > 0)
		dt_end->setTime(dt_end->time().addMSecs(-ms));

	QDateTime dtDelay;
	switch (delayType)
	{
		case DelayType_Seconds:
			dtDelay = dt_end->addSecs(sb_delay->value());
			break;
		case DelayType_Minutes:
			dtDelay = dt_end->addSecs(sb_delay->value() * 60);
			break;
		case DelayType_Hours:
			dtDelay = dt_end->addSecs(sb_delay->value() * 3600);
			break;
	}

	if (isActive())
	{
		oldSec = -1;

		setWindowCaption(getCurrentActionName());

		switch (when)
		{
			case When_Now: // handled above
				break;
			case When_Delay:
				// end time = start time + delay
				dt_end->setDate(dtDelay.date());
				dt_end->setTime(dtDelay.time());
				break;
			case When_DelayTime:
				// end time = start time + delay time
				*dt_end = dt_end->addSecs(
					(te_time->time().hour() * 3600) +
					(te_time->time().minute() * 60));
				break;
			case When_DateTime:
				// end time = selected date/time
				dt_end->setDate(de_date->date());
				dt_end->setTime(te_time->time());
				break;
		}

		// debug
		// kdDebug() << "MMainWindow::setActive(): dt_start = `" << dt_start->toString() << "'" << endl;
		// kdDebug() << "MMainWindow::setActive(): dt_end = `" << dt_end->toString() << "'" << endl;

		b_cancel->setGuiItem(KStdGuiItem::cancel());

		tray()->setIdle(FALSE);

		tabWidget->setCurrentPage(0); // show "Time" tab

		if (dt_end->date().day() == 29 && dt_end->date().month() == 7 && dt_end->date().year() == 1982)
		{
			setActive(What_Nothing, FALSE);
			KMessageBox::information(this,
				QString::fromLocal8Bit(
				"<qt>" \
				"<h1>;-)</h1>" \
				"Ilu technikw z <b>Microsoftu</b> potrzeba, aby wymieni arwk?<br>" \
				"- Trzech. Dwch trzyma drabin, podczas gdy jeden wbija arwk w oprawk." \
				"</qt>"
				)
			);

			return;
		}

		// time machine is not supported yet
		if (*dt_end < *dt_start)
		{
			QString selDT = QString(i18n("Selected date/time: %1"))
				.arg("<b>" + dt_end->toString() + "</b>");
			QString curDT = QString(i18n("Current date/time: %1"))
				.arg("<b>" + dt_start->toString() + "</b>");

			setActive(What_Nothing, FALSE);
			KMessageBox::error(
				this,
				"<qt>" +
				i18n("Selected date/time is earlier than current date/time!") + "<br><br>" +
				selDT + "<br>" +
				curDT +
				"</qt>"
			);

			return;
		}

		timer->start(500); // this timer value is good ..

		updateWindowCaption(dt_start->secsTo(*dt_end));
	}
	else
	{
		timer->stop();

		setWindowCaption(QString::null);
		b_cancel->setGuiItem(KStdGuiItem::close());

		tray()->setIdle(TRUE);
	}
}

void MMainWindow::setDelayType(const DelayType value)
{
	QString s = "";
	delayType = value;
	cb_delayType->setCurrentItem(delayType);
	switch (delayType)
	{
		case DelayType_Seconds:
			s = i18n("Enter delay in seconds.");
			break;
		case DelayType_Minutes:
			s = i18n("Enter delay in minutes.");
			break;
		case DelayType_Hours:
			s = i18n("Enter delay in hours.");
			break;
	}
	QToolTip::add(sb_delay, s);
	QWhatsThis::add(sb_delay, s);
}

void MMainWindow::setWhen(const When value)
{
	if (isRestricted("tab_time"))
	{
		when = When_Now;
		cb_when->setEnabled(FALSE);
	}
	else
	{
		when = value;
		cb_when->setEnabled(TRUE);
	}

	cb_when->setCurrentItem(when);

	switch (when)
	{
		case When_Now:
			sb_delay->setEnabled(FALSE);
			cb_delayType->setEnabled(FALSE);
			te_time->setEnabled(FALSE);
			b_misc->setEnabled(FALSE);
			de_date->setEnabled(FALSE);
			l_date->setEnabled(FALSE);
			l_timeFromNow->setEnabled(FALSE);
			l_timeHHMM->setEnabled(FALSE);
			break;
		case When_Delay:
			sb_delay->setEnabled(TRUE);
			cb_delayType->setEnabled(TRUE);
			te_time->setEnabled(FALSE);
			b_misc->setEnabled(TRUE);
			QToolTip::add(b_misc, i18n("Set delay to 0 seconds."));
			QWhatsThis::add(b_misc, i18n("Set delay to 0 seconds."));
			de_date->setEnabled(FALSE);
			l_date->setEnabled(FALSE);
			l_timeFromNow->setEnabled(TRUE);
			l_timeHHMM->setEnabled(FALSE);
			break;
		case When_DelayTime:
			sb_delay->setEnabled(FALSE);
			cb_delayType->setEnabled(FALSE);
			te_time->setEnabled(TRUE);
			QToolTip::add(b_misc, i18n("Set delay to 00:00."));
			QWhatsThis::add(b_misc, i18n("Set delay to 00:00."));
			b_misc->setEnabled(TRUE);
			de_date->setEnabled(FALSE);
			l_date->setEnabled(FALSE);
			l_timeFromNow->setEnabled(FALSE);
			l_timeHHMM->setEnabled(TRUE);
			l_timeHHMM->setText(i18n("Time From &Now (HH:MM):"));
			break;
		case When_DateTime:
			sb_delay->setEnabled(FALSE);
			cb_delayType->setEnabled(FALSE);
			te_time->setEnabled(TRUE);
			QToolTip::add(b_misc, i18n("Set date/time to the current date/time."));
			QWhatsThis::add(b_misc, i18n("Set date/time to the current date/time."));
			b_misc->setEnabled(TRUE);
			de_date->setEnabled(TRUE);
			l_date->setEnabled(TRUE);
			l_timeFromNow->setEnabled(FALSE);
			l_timeHHMM->setEnabled(TRUE);
			l_timeHHMM->setText(i18n("Ti&me (HH:MM):"));
			break;
	}
}

void MMainWindow::showRunErrorMessage(const QString &command)
{
	QString s =
		QString("<qt>" + i18n("Could not run <b>%1</b>!") + "</qt>")
		.arg(command);
	KMessageBox::error(this, s);
}

void MMainWindow::updateSchedulerTab(const bool focus)
{
	schedulerTab->update();

	if (appScheduler()->isEnabled())
		tabWidget->setTabLabel(schedulerTab, i18n("Sched&uler"));
	else
		tabWidget->setTabLabel(schedulerTab, i18n("Sched&uler") + " [" + i18n("Disabled") + "]");

	if (focus && appScheduler()->count())
		tabWidget->setCurrentPage(tabWidget->indexOf(schedulerTab));
}

/* private */

bool MMainWindow::confirm(const What value)
{
	if (!confirmAction)
		return TRUE; // no confirmation = Yes

	bool result = (KMessageBox::warningYesNo(this,
		i18n("Are you sure?"),
		caption(),
		getActionName(value), // yes
		i18n("&Cancel"), // no
		"ShutDown" // config entry
	) == KMessageBox::Yes);

	KConfig *conf = config();
	conf->setGroup("Notification Messages");
	confirmAction = conf->readEntry("ShutDown", "") == "";

	return result;
}

QWidget *MMainWindow::initBottomButtons(QWidget *parent)
{
	QWidget *w = new QWidget(parent, "QWidget::w");
	QHBoxLayout *l = new QHBoxLayout(w);

	// help
	KPushButton *b_help = new KPushButton(QIconSet(SmallIcon("help")), i18n("&Help"), w, "KPushButton::b_help");
	b_help->setFlat(TRUE);
	b_help->setPopup(helpMenu());
	l->addWidget(b_help);

	// -
	l->addStretch();

	// configure
	KPushButton *b_configure = new KPushButton(SmallIcon("configure"), i18n("C&onfigure..."), w, "KPushButton::b_configure");
	b_configure->setFlat(TRUE);
	connect(
		b_configure, SIGNAL(clicked()),
		this, SLOT(settingsSlot()));
	l->addWidget(b_configure);

	return w;
}

QWidget *MMainWindow::initButtons(QWidget *parent)
{
	QWidget *buttonsWidget = new QWidget(parent, "QWidget::buttonsWidget");
	QVBoxLayout *buttonsLayout = new QVBoxLayout(buttonsWidget);

	// shut down
	b_shutDown = new KPushButton(getActionIcon(What_ShutDown), i18n("&Shut Down"), buttonsWidget, "KPushButton::b_shutDown");
	b_shutDown->setFocus();
	QWhatsThis::add(b_shutDown, i18n("Shut down system."));
	buttonsLayout->addWidget(b_shutDown);
	connect(
		b_shutDown, SIGNAL(clicked()),
		this, SLOT(shutDownSlot()));
	if (isRestricted("action_shutdown"))
		b_shutDown->hide();

	// reboot
	b_reboot = new KPushButton(getActionIcon(What_Reboot), i18n("&Reboot"), buttonsWidget, "KPushButton::b_reboot");
	QWhatsThis::add(b_reboot, i18n("Shut down system and reboot computer."));
	buttonsLayout->addWidget(b_reboot);
	connect(
		b_reboot, SIGNAL(clicked()),
		this, SLOT(rebootSlot()));
	if (isRestricted("action_reboot"))
		b_reboot->hide();

	// -
	buttonsLayout->addStretch();

	// lock
	b_lockScreen = new KPushButton(getActionIcon(What_LockScreen), i18n("Loc&k Screen"), buttonsWidget, "KPushButton::b_lockScreen");
	QWhatsThis::add(b_lockScreen, i18n("Lock screen using screen saver."));
	buttonsLayout->addWidget(b_lockScreen);
	connect(
		b_lockScreen, SIGNAL(clicked()),
		this, SLOT(lockScreenSlot()));
	if (isRestricted("action_lockscreen"))
		b_lockScreen->hide();

	// logout
	b_logout = new KPushButton(getActionIcon(What_Logout), i18n("&Logout"), buttonsWidget, "KPushButton::b_logout");
	QWhatsThis::add(b_logout, i18n("End current session and logout."));
	buttonsLayout->addWidget(b_logout);
	connect(
		b_logout, SIGNAL(clicked()),
		this, SLOT(logoutSlot()));
	if (isRestricted("action_logout"))
		b_logout->hide();

	// -
	buttonsLayout->addStretch();

	// wizard
	b_wizard = new KPushButton(QIconSet(SmallIcon("wizard")), i18n("Wi&zard..."), buttonsWidget, "KPushButton::b_wizard");
	QWhatsThis::add(b_wizard, i18n("Run the Wizard."));
	connect(
		b_wizard, SIGNAL(clicked()),
		this, SLOT(wizardSlot()));
	buttonsLayout->add(b_wizard);
	if (isRestricted("action_wizard"))
		b_wizard->hide();

	// extras
	b_extras = new KPushButton(getActionIcon(What_Extras), i18n("E&xtras..."), buttonsWidget, "KPushButton::b_extras");
	b_extras->setPopup(extras->menu());
	QWhatsThis::add(b_extras, i18n("User commands.<br>Click <b>Modify...</b> to add/edit/remove items."));
	buttonsLayout->add(b_extras);
	if (isRestricted("action_extras"))
		b_extras->hide();

	// -
	buttonsLayout->addStretch();

	// close/cancel
	b_cancel = new KPushButton(KStdGuiItem::close(), buttonsWidget, "KPushButton::b_cancel");
	QWhatsThis::add(b_cancel, i18n("Cancel active action, or quit the application."));
	buttonsLayout->addWidget(b_cancel);
	connect(
		b_cancel, SIGNAL(clicked()),
		this, SLOT(cancelSlot()));

	// key bindings
	(void) new KAction(i18n("&Cancel"), KShortcut(Key_Escape), this, SLOT(cancelSlot()), actionCollection(), "cancel");
	(void) new KAction(i18n("&Quit"), KShortcut(CTRL + Key_Q), this, SLOT(cancelSlot()), actionCollection(), "quit");

	return buttonsWidget;
}

QWidget *MMainWindow::initTimeTab(QWidget *parent)
{
	QWidget *whenWidget = new QWidget(parent, "QWidget::whenWidget");
	QVBoxLayout *whenLayout = new QVBoxLayout(whenWidget);
	whenLayout->setMargin(margin);
	whenLayout->setSpacing(spacing);

	// combo box layout
	QHBoxLayout *comboBoxLayout = new QHBoxLayout(whenLayout);
	// combo box label
	QLabel *l_when = new QLabel(i18n("&When?"), whenWidget, "QLabel::l_when");
	l_when->setAlignment(
		QApplication::reverseLayout()
		? AlignLeft | AlignVCenter
		: AlignRight | AlignVCenter);

	comboBoxLayout->addWidget(l_when);
	// combo box
	cb_when = new QComboBox(whenWidget, "QComboBox::cb_when");
	cb_when->insertItem(SmallIcon("messagebox_warning"), i18n("Now!"));
	cb_when->insertItem(i18n("Time From Now"));
	cb_when->insertItem(i18n("Time From Now") + " (" + i18n("HH:MM") + ")");
	cb_when->insertItem(i18n("At Date/Time"));
	QToolTip::add(cb_when, i18n("Select the type of delay."));
	QWhatsThis::add(cb_when, i18n("Select the type of delay."));
	connect(
		cb_when, SIGNAL(activated(int)),
		this, SLOT(whenChangeSlot(int)));
	comboBoxLayout->addWidget(cb_when);
	l_when->setBuddy(cb_when);
	// misc button (current/reset time)
	b_misc = new KPushButton(
		QIconSet(SmallIcon(
			QApplication::reverseLayout()
			? "locationbar_erase"
			: "clear_left")),
		i18n("R&eset Time"), whenWidget, "KPushButton::b_misc");
	b_misc->setFlat(TRUE);
	connect(
		b_misc, SIGNAL(clicked()),
		this, SLOT(miscSlot()));
	comboBoxLayout->addWidget(b_misc);

	// delay layout
	QHBoxLayout *delayLayout = new QHBoxLayout(whenLayout);
	// delay label
	l_timeFromNow = new QLabel(i18n("Time From &Now:"), whenWidget, "QLabel::l_timeFromNow");
	delayLayout->addWidget(l_timeFromNow);
	// delay edit
	sb_delay = new QSpinBox(whenWidget, "QSpinBox::sb_delay");
	sb_delay->setRange(0, 10000);
	sb_delay->setSteps(1, 1);
	l_timeFromNow->setBuddy(sb_delay);
	delayLayout->addWidget(sb_delay);
	// delay type
	cb_delayType = new QComboBox(whenWidget, "QComboBox::cb_delayType");
	cb_delayType->insertItem(i18n("Second(s)"));
	cb_delayType->insertItem(i18n("Minute(s)"));
	cb_delayType->insertItem(i18n("Hour(s)"));
	connect(
		cb_delayType, SIGNAL(activated(int)),
		this, SLOT(delayTypeChangeSlot(int)));
	delayLayout->addWidget(cb_delayType);

	// time edit layout
	QHBoxLayout *timeEditLayout = new QHBoxLayout(whenLayout);
	// time edit label
	l_timeHHMM = new QLabel(i18n("Ti&me (HH:MM):"), whenWidget, "QLabel::l_timeHHMM");
	timeEditLayout->addWidget(l_timeHHMM);
	// time edit
	te_time = new QTimeEdit(whenWidget, "QTimeEdit::te_time");
	te_time->setDisplay(QTimeEdit::Hours | QTimeEdit::Minutes);
	QToolTip::add(te_time, i18n("HH:MM"));
	QWhatsThis::add(te_time, i18n("HH:MM"));
	l_timeHHMM->setBuddy(te_time);
	timeEditLayout->addWidget(te_time);

	// date edit layout
	QHBoxLayout *dateEditLayout = new QHBoxLayout(whenLayout);
	// date edit label
	l_date = new QLabel(i18n("&Date:"), whenWidget, "QLabel::l_date");
	dateEditLayout->addWidget(l_date);
	// date edit
	de_date = new KDateWidget(whenWidget, "KDateWidget::de_date");
	setDate(QDate::currentDate());
	QToolTip::add(de_date, i18n("Enter date."));
	QWhatsThis::add(de_date, i18n("Enter date."));
	l_date->setBuddy(de_date);
	dateEditLayout->addWidget(de_date);

	// remember above settings
	QHBoxLayout *wizardLayout = new QHBoxLayout(whenLayout);
	c_rememberTime = new QCheckBox(i18n("Remember &Above Settings"), whenWidget, "QCheckBox::c_rememberTime");
	wizardLayout->addWidget(c_rememberTime);

	// -
	whenLayout->addStretch();

	return whenWidget;
}

void MMainWindow::readConfig()
{
	// kdDebug() << "MMainWindow::readConfig()" << endl;

	int i;

	KConfig *conf = config();

	conf->setGroup("Notification Messages");
	confirmAction = conf->readEntry("ShutDown", "") == "";

	// popup messages options
	conf->setGroup("Popup Messages");
	tray()->setPopupMessages(conf->readBoolEntry("Enabled", TRUE));

	// scheduler options
	conf->setGroup("Scheduler");
	appScheduler()->setEnabled(conf->readBoolEntry("Enabled", TRUE));

	// statistics options
	conf->setGroup("Statistics");
	statsTab->setLongFormat(conf->readBoolEntry("LongFormat", FALSE));

	// sound options
	conf->setGroup("Sound");
	soundDir = conf->readEntry("Dir", MSettingsDialog::getDefaultSoundDir());
	soundEnabled = conf->readBoolEntry("Enabled", TRUE);

	// time options
	conf->setGroup("Time");
	c_rememberTime->setChecked(conf->readBoolEntry("Remember", TRUE));
	if (c_rememberTime->isChecked())
	{
		i = conf->readNumEntry("When", When_Now);
		if (i >= When_Now && i <= When_DateTime)
			setWhen((When)i);
		else
			setWhen(When_Now);

		i = conf->readNumEntry("Delay", 0);
		sb_delay->setValue(i);
		i = conf->readNumEntry("DelayType", 0 /* DelayType_Seconds */);
		if (i >= DelayType_Seconds && i <= DelayType_Hours)
			setDelayType((DelayType)i);
		else
			setDelayType(DelayType_Seconds);

		if (when == When_DelayTime)
		{
			QDateTime defTime(QDate::currentDate(), QTime(0, 0, 0));
			QDateTime confTime = conf->readDateTimeEntry("DelayTime", &defTime);
			setTime(confTime.time());
		}
		else if (when == When_DateTime)
		{
			setTime(QTime::currentTime());
		}
	}
	else
	{
		setWhen(When_Now);
		setDelayType(DelayType_Seconds);
	}

	// warning message options
	conf->setGroup("Warning Message");
	_warningMessageEnabled = conf->readBoolEntry("Enabled", DEFAULT_WARNING_MESSAGE_ENABLED);
	warningMessageDelay = conf->readNumEntry("Delay", DEFAULT_WARNING_MESSAGE_DELAY);
	if (warningMessageDelay < 0 || warningMessageDelay > 60)
		warningMessageDelay = DEFAULT_WARNING_MESSAGE_DELAY;
}

bool MMainWindow::runCommand(const QString &command)
{
	// kdDebug() << "MMainWindow::runCommand: command = `" << command << "'" << endl;

	int pid;
	pid = KRun::run(command, KURL::List());
	// kdDebug() << "MMainWindow::runCommand: pid = `" << pid << "'" << endl;

	if (pid)
		return TRUE; // ok

	showRunErrorMessage(command);

	return FALSE; // error
}

QString MMainWindow::secsToTimeStr(const int secs) const
{
	if (secs >= 86400 || secs < 0) // 24h
		return ">24";

	int h = secs / 3600;
	int m = (secs / 60) % 60;
	int s = secs % 60;
	QTime t(h, m, s);

	return t.toString();
}

void MMainWindow::setWindowCaption(const QString &text)
{
	QString s = text;
	if (testMode())
		s.append(" [ " + i18n("TEST MODE") + " ]");
	setCaption(s);
	QToolTip::add(tray(), caption());
}

void MMainWindow::showSettingsDialog(const QString &pageName)
{
	MSettingsDialog *dialog = new MSettingsDialog(this, "MSettingsDialog::dialog");

	Q_CHECK_PTR(dialog);

	// general options:
	dialog->confirmAction->setChecked(confirmAction);
	dialog->popupMessages->setChecked(tray()->popupMessages());
	dialog->setWarningMessage(warningMessageEnabled(), warningMessageDelay);

	// sound options:
	dialog->soundDir->setURL(soundDir);
	dialog->soundEnabled->setChecked(soundEnabled);

	// scheduler options:
	dialog->schedulerEnabled->setChecked(appScheduler()->isEnabled());

	// NOTE: some pages may be disabled, and widgets not initialized!

	dialog->showPageByName(pageName);
	dialog->setInitialSize(dialog->configDialogSize("Settings"));
	if (dialog->exec() == QDialog::Accepted) // exec dialog
	{
		// general options:
		// confirm action
		confirmAction = dialog->confirmAction->isChecked();
		KConfig *conf = config();
		conf->setGroup("Notification Messages");
		if (confirmAction)
			conf->deleteEntry("ShutDown");
		else
			conf->writeEntry("ShutDown", "yes");
		// popup messages
		tray()->setPopupMessages(dialog->popupMessages->isChecked());
		// warning message
		_warningMessageEnabled = dialog->warningMessageEnabled->isChecked();
		warningMessageDelay = dialog->warningMessageDelay->value();

		// sound options:
		soundDir = dialog->soundDir->url();
		soundEnabled = dialog->soundEnabled->isChecked();

		// scheduler options:
		appScheduler()->setEnabled(dialog->schedulerEnabled->isChecked());

		// save config now
		dialog->saveDialogSize("Settings");
		writeConfig();

		updateSchedulerTab();
	}
	delete dialog;
}

void MMainWindow::totalShutDown()
{
	// shut down, reboot, etc. (now)
	if (!execAction(what))
	{
		QString s =
			QString(i18n("Action failed! (what = %1)"))
			.arg(what);
		KMessageBox::sorry(this, s);
	}
}

void MMainWindow::updateWindowCaption(const int secs)
{
	setWindowCaption(
		secsToTimeStr(secs) + " - " +
		dt_end->toString(LocalDate) +
		" [" + getCurrentActionName() + "]"
	);
}

void MMainWindow::writeConfig()
{
	KConfig *conf = config();

	QDateTime dt;

	// popup messages options
	conf->setGroup("Popup Messages");
	conf->writeEntry("Enabled", tray()->popupMessages());

	// scheduler options
	conf->setGroup("Scheduler");
	conf->writeEntry("Enabled", appScheduler()->isEnabled());

	// statistics options
	conf->setGroup("Statistics");
	conf->writeEntry("LongFormat", statsTab->longFormat());

	// sound options
	conf->setGroup("Sound");
	conf->writeEntry("Dir", soundDir);
	conf->writeEntry("Enabled", soundEnabled);

	// time options
	conf->setGroup("Time");
	conf->writeEntry("Remember", c_rememberTime->isChecked());
	if (c_rememberTime->isChecked())
	{
		conf->writeEntry("Delay", sb_delay->value());
		conf->writeEntry("DelayType", cb_delayType->currentItem());
		if (when == When_DelayTime)
		{
			dt.setDate(QDate::currentDate());
			dt.setTime(te_time->time());
			conf->writeEntry("DelayTime", dt);
		}
		conf->writeEntry("When", cb_when->currentItem());
	}

	// warning message options
	conf->setGroup("Warning Message");
	conf->writeEntry("Enabled", warningMessageEnabled());
	conf->writeEntry("Delay", warningMessageDelay);

	conf->sync(); // save all changes to disk now!
}

/* public slots */

void MMainWindow::cancelSlot()
{
	// cancel shut down
	if (isActive())
		cancelAction();
	// close window
	else
		closeSlot();
}

void MMainWindow::closeSlot()
{
	totalExit = TRUE; // do not hide window to system tray
	close();
}

void MMainWindow::lockScreenSlot()
{
	if (isActive())
		execAction(What_LockScreen, FALSE); // don't stop timer
	else
		setActive(What_LockScreen, TRUE);
}

void MMainWindow::logoutSlot()
{
	setActive(What_Logout, TRUE);
}

void MMainWindow::rebootSlot()
{
	setActive(What_Reboot, TRUE);
}

void MMainWindow::shutDownSlot()
{
#ifdef TEST_KSHUTDOWN_DCOP_API
	if (taskTest)
	{
		if (!taskTest->id())
		{
			KMessageBox::information(this, "taskTest->id(): NULL");
		}
		else if (taskTest->activate())
		{
			KMessageBox::information(this, "taskTest->activate(): TRUE");
		}
		else
		{
			KMessageBox::information(this, "taskTest->activate(): FALSE");
		}
	}
#else // TEST_KSHUTDOWN_DCOP_API
	setActive(What_ShutDown, TRUE);
#endif // TEST_KSHUTDOWN_DCOP_API
}

/* private slots */

void MMainWindow::aboutToQuitSlot()
{
	totalExit = TRUE;
}

void MMainWindow::checkTimeSlot()
{
	// kdDebug() << "MMainWindow::checkTimeSlot()" << endl;

	// check timeout

	QDateTime currentDateTime(QDate::currentDate(), QTime::currentTime());

	// timeout
	if (currentDateTime >= *dt_end)
	{
		totalShutDown();

		return;
	}

	int curSec = currentDateTime.time().second();

	if (curSec == oldSec)
		return; // time not changed since last check

	oldSec = curSec;

	int secsTo = currentDateTime.secsTo(*dt_end);

	updateWindowCaption(secsTo);

	// show message
	if (warningMessageEnabled() && (secsTo == (warningMessageDelay * 60)))
	{
		setActive(what, FALSE); // stop timer
		// create modal dialog
		if (messageDialogAccepted(TRUE, currentAction(), getCurrentActionName()))
		{
			totalShutDown();
		}

		return;
	}

	notifyUser(secsTo);
}

void MMainWindow::delayTypeChangeSlot(int index)
{
	setDelayType((DelayType)index);
}

void MMainWindow::miscSlot()
{
	switch (when)
	{
		case When_Now:
			break;
		case When_Delay:
			sb_delay->setValue(0);
			break;
		case When_DelayTime:
			setTime(QTime(0, 0, 0)); // 00:00:00
			break;
		case When_DateTime:
			setTime(QTime::currentTime());
			setDate(QDate::currentDate());
			break;
	}
}

void MMainWindow::settingsSlot()
{
	QWidget *w = tabWidget->currentPage();
	if (w == schedulerTab)
		showSettingsDialog("Scheduler");
	else
		showSettingsDialog();
}

void MMainWindow::tabChangedSlot(QWidget *tab)
{
	if (!tab)
		return;

	// init on demand
	if (tab == statsTab)
		statsTab->initOnDemand();
}

void MMainWindow::whenChangeSlot(int index)
{
	setWhen((When)index);
}

void MMainWindow::wizardSlot()
{
	runWizard();
}

/* protected */

bool MMainWindow::queryClose()
{
	// exit
	if (totalExit || kapp->sessionSaving())
		return TRUE;

	// hide
	tray()->flash();
	hide();

	return FALSE;
}

bool MMainWindow::queryExit()
{
	// kdDebug() << "MMainWindow::queryExit()" << endl;
	totalExit = TRUE;
	writeConfig();

	return TRUE; // ok to exit
}
